@@ -1,10 +1,9 @@
 from superpowered.main import BASE_URL, _format_http_response, get_headers
 import requests
-import random
-import string
 import os
-#from llama_index import download_loader
-from pathlib import Path
+import PyPDF2
+import docx
+from bs4 import BeautifulSoup
 
 class KnowledgeBase:
     """
@@ -146,7 +145,6 @@ def get_knowledge_base(title: str):
             kb_from_title[kb.title] = kb
         else:
             print ('WARNING: Duplicate knowledge base title: ' + kb.title)
-            #raise Exception('Duplicate knowledge base title: ' + kb.title)
     
     # if the title we're looking for is in the map, return the KnowledgeBase object
     if title in kb_from_title:
@@ -177,66 +175,139 @@ def list_documents_in_kb(kb_title: str, verbose=True):
             print (f"title: {document_obj.title}\ncontent: {document_obj.content[:100]}\n\n")
     return documents
 
-def read_file(file_path: str, file_type: str = None):
+def read_file(file_path: str):
     """
     read a file and return its contents as a string
     """
-    if file_type is None:
-        file_type = file_path.split(".")[-1]
+    # infer the file type from the file extension
+    file_type = file_path.split(".")[-1]
 
-    reader_names = {
-        "pdf": "PDFReader",
-        "csv": "SimpleCSVReader",
-        "docx": "DocxReader",
-        "md": "MarkdownReader"
-    }
+    if file_type == "pdf":
+        # Open the PDF file
+        with open(file_path, 'rb') as file:
+            # Create a PDF reader object
+            pdf_reader = PyPDF2.PdfReader(file)
 
-    """elif file_type in list(reader_names.keys()):
-        Reader = download_loader(reader_names[file_type])
-        loader = Reader()
-        documents = loader.load_data(file=Path(file_path))
-        content = documents[0].text
-        return content"""
-    
-    if file_type == "txt":
+            # Initialize an empty string to store the extracted text
+            extracted_text = ""
+
+            # Loop through all the pages in the PDF file
+            for page_num in range(len(pdf_reader.pages)):
+                # Extract the text from the current page
+                page_text = pdf_reader.pages[page_num].extract_text()
+
+                # Add the extracted text to the final text
+                extracted_text += page_text 
+        return extracted_text
+    elif file_type == "txt":
         with open(file_path, 'r') as f:
             # read the file and convert it to a string
             content = f.read()
-            return content
-    else:
-        return None
+        return content
+    elif file_type == "docx":
+        # Create a docx Document object
+        doc = docx.Document(file_path)
 
-def add_file_to_kb(kb_title: str, file_path: str, file_type: str = None):
+        # Initialize an empty string to store the extracted text
+        extracted_text = ""
+
+        # Loop through all the paragraphs in the docx file
+        for paragraph in doc.paragraphs:
+            # Add the extracted text to the final text
+            extracted_text += paragraph.text + "\n"
+
+        return extracted_text
+    elif file_type == "md":
+        with open(file_path, 'r') as f:
+            # read the file and convert it to a string
+            content = f.read()
+        return content
+    else:
+        raise Exception("File type not supported. Supported file types are: .txt, .pdf, .docx, and .md")
+    
+def add_document_to_kb(kb_title: str, content: str, title: str = None, link_to_source: str = None, supp_id: str = None, description: str = None):
     """
-    add a file to a knowledge base - currently supports .txt, .pdf, .csv, .docx, and .md files
+    add a document to a knowledge base by providing the title and content (as a string)
+    - use this function if you want to directly add text to a knowledge base. If you want to add a file, use add_file_to_kb() instead.
     """
     kb = get_knowledge_base(kb_title)
-    if file_type is None:
-        file_type = file_path.split(".")[-1]
-    
-    content = read_file(file_path, file_type)
+    kb.add_document(content=content, title=title, link_to_source=link_to_source, supp_id=supp_id, description=description)
 
-    if content is None:
-        raise Exception("File type not supported: " + file_type)
+def add_file_to_kb(kb_title: str, file_path: str, title: str = None, link_to_source: str = None, supp_id: str = None, description: str = None):
+    """
+    add a Document to a knowledge base by reading the contents of a file - currently supports .txt, .pdf, .docx, and .md files
+    """
+    content = read_file(file_path)
     
-    # make the title a random string
-    title = ''.join(random.choices(string.ascii_uppercase + string.digits, k=10))
-    kb.add_document(content=content, title=title)
+    # make the title the file name if it's not provided
+    if title is None:
+        title = file_path.split("/")[-1]
+    
+    kb = get_knowledge_base(kb_title)
+    kb.add_document(content=content, title=title, link_to_source=link_to_source, supp_id=supp_id, description=description)
 
 def add_directory_to_kb(kb_title: str, directory_path: str, verbose=False):
     """
     add all the supported files in a directory to a knowledge base
     """
-    # TODO: add support for other file types
-    # TODO: make sure this works for all different kinds of subdirectory structures - currently it's only been tested for one level of subdirectories
-    # get all the txt files in the directory, and all subdirectories
+    # get a list of all the files in the directory
     all_files = []
     for root, _, files in os.walk(directory_path):
         for file in files:
-            if file.endswith(".txt"):
+            if file.endswith(".txt") or file.endswith(".pdf") or file.endswith(".docx") or file.endswith(".md"):
                 all_files.append(os.path.join(root, file))
 
     # add each file to the knowledge base
     for file in all_files:
         if verbose: print("Adding:", file)
-        add_file_to_kb(kb_title, file_path=file)
\ No newline at end of file
+        add_file_to_kb(kb_title, file_path=file)
+
+def add_url_to_kb(kb_title: str, url: str, title: str = None, link_to_source: str = None, supp_id: str = None, description: str = None):
+    """
+    add a Document to a knowledge base by scraping the contents of a URL
+    """
+    # scrape the URL
+    response = requests.get(url)
+    soup = BeautifulSoup(response.content, "html.parser")
+    body = soup.find('body')
+    content = body.text
+
+    # make the title the URL if no title is provided
+    if title is None:
+        title = url
+
+    kb = get_knowledge_base(kb_title)
+    kb.add_document(content=content, title=title, link_to_source=link_to_source, supp_id=supp_id, description=description)
+
+def query(query: str, kb_titles: list = [], kb_ids: list = [], retriever_top_k: int = 100, reranker_top_k: int = 5, extract_and_summarize: bool = False):
+    """
+    Can take either a list of kb_titles or a list of kb_ids (or a mix of both)
+    """
+    # convert kb_titles to kb_ids
+    if len(kb_titles) > 0:
+        knowledge_bases = get_knowledge_bases() # gets a dict of KnowledgeBase objects keyed on kb_id
+        for kb_title in kb_titles:
+            # see if the title is in the knowledge bases dict
+            for kb_id in knowledge_bases.keys():
+                if knowledge_bases[kb_id].title == kb_title:
+                    kb_ids.append(kb_id)
+                    break
+
+    # deduplicate kb_ids
+    kb_ids = list(set(kb_ids))
+
+    # make sure kb_ids is not empty
+    if len(kb_ids) == 0:
+        raise Exception("Must provide at least one valid knowledge base title or id")
+
+    # make the request
+    url = BASE_URL + f'knowledge_bases/query'
+    payload = {
+        'query': query,
+        'knowledge_base_ids': kb_ids,
+        'retriever_top_k': retriever_top_k,
+        'reranker_top_k': reranker_top_k,
+        'summarize_results': extract_and_summarize,
+    }
+    resp = _format_http_response(requests.post(url, headers=get_headers(), json=payload))
+    return resp["body"]
\ No newline at end of file
